@@ -27,8 +27,11 @@ from segments.exceptions import (
     AlreadyExistsError,
     APILimitError,
     AuthenticationError,
+    AuthorizationError,
+    CollaboratorError,
     NetworkError,
     NotFoundError,
+    SubscriptionError,
     TimeoutError,
     ValidationError,
 )
@@ -53,20 +56,21 @@ from segments.typing import (
     TaskAttributes,
     TaskType,
 )
-from typing_extensions import Literal
+from typing_extensions import Literal, get_args
 
-#############
-# Variables #
-#############
+################################
+# Constants and type variables #
+################################
 logger = logging.getLogger(__name__)
 T = TypeVar("T")
+VERSION = "1.0.17"
 
 
 ####################
 # Helper functions #
 ####################
 # Error handling: https://stackoverflow.com/questions/16511337/correct-way-to-try-except-using-python-requests-module
-def exception_handler(
+def handle_exceptions(
     f: Callable[..., requests.Response]
 ) -> Callable[..., Union[requests.Response, T]]:
     """Catch exceptions and throw Segments exceptions.
@@ -82,45 +86,48 @@ def exception_handler(
         :exc:`~segments.exceptions.TimeoutError`: If the request times out - catches :exc:`requests.exceptions.TimeoutError`.
     """
 
-    def wrapper_function(
+    def throw_segments_exception(
         *args: Any,
         model: Optional[Type[T]] = None,
         **kwargs: Any,
     ) -> Union[requests.Response, T]:
         try:
-            self = args[0]
             r = f(*args, **kwargs)
             r.raise_for_status()
             if r.content:
                 r_json = r.json()
                 # Check if the API limit is exceeded
                 if isinstance(r_json, dict):
-                    message = r_json.get("message", "")
-                    if message.startswith("You have exceeded"):
+                    message = r_json.get("detail", "")
+                    if "throttled" in message:
                         raise APILimitError(message)
                 if model:
-                    if self.disable_typechecking:
-                        try:
-                            if isinstance(model.__origin__(), list):
-                                model_without_list = model.__args__[0]
-                                m = [
-                                    model_without_list.construct(r_j) for r_j in r_json
-                                ]
-                        except AttributeError:
-                            m = model.construct(r_json)
-                    else:
-                        m = parse_obj_as(model, r_json)
+                    m = parse_obj_as(model, r_json)
                     return m
             return r
         except requests.exceptions.Timeout as e:
             # Maybe set up for a retry, or continue in a retry loop
             raise TimeoutError(message=str(e), cause=e)
         except requests.exceptions.HTTPError as e:
-            text = e.response.text
-            if "Not found" in text:
+            # Make string comparison case insensitive.
+            text = e.response.text.lower()
+            if "not found" in text or "does not exist" in text:
                 raise NotFoundError(message=text, cause=e)
-            if "already exists" in text:
+            if "already exists" in text or "already have" in text:
                 raise AlreadyExistsError(message=text, cause=e)
+            if (
+                "cannot be added as collaborator" in text
+                or "is already a collaborator" in text
+            ):
+                raise CollaboratorError(message=text, cause=e)
+            if (
+                "cannot leave the organization" in text
+                or "need to be an administrator" in text
+                or "do not have permission" in text
+            ):
+                raise AuthorizationError(message=text, cause=e)
+            if "free trial ended" in text or "exceeded user limit" in text:
+                raise SubscriptionError(message=text, cause=e)
             raise NetworkError(message=text, cause=e)
         except requests.exceptions.TooManyRedirects as e:
             # Tell the user their URL was bad and try a different one
@@ -131,7 +138,7 @@ def exception_handler(
         except pydantic.ValidationError as e:
             raise ValidationError(message=str(e), cause=e)
 
-    return wrapper_function
+    return throw_segments_exception
 
 
 ##########
@@ -177,7 +184,6 @@ class SegmentsClient:
         self,
         api_key: Optional[str] = None,
         api_url: str = "https://api.segments.ai/",
-        disable_typechecking: bool = False,
     ):
         if api_key is None:
             api_key = os.getenv("SEGMENTS_API_KEY")
@@ -190,7 +196,6 @@ class SegmentsClient:
 
         self.api_key = api_key
         self.api_url = api_url
-        self.disable_typechecking = disable_typechecking
 
         # https://realpython.com/python-requests/#performance
         # https://stackoverflow.com/questions/21371809/cleanly-setting-max-retries-on-python-requests-get-or-post-method
@@ -205,7 +210,7 @@ class SegmentsClient:
         self.s3_session.mount("https://", adapter)
 
         try:
-            r = self._get("/api_status/?lib_version=1.0.11")
+            r = self._get(f"/api_status/?lib_version={VERSION}")
             if r.status_code == 200:
                 logger.info("Initialized successfully.")
         except NetworkError as e:
@@ -321,6 +326,8 @@ class SegmentsClient:
         enable_skip_reviewing: bool = False,
         enable_ratings: bool = False,
         enable_interpolation: bool = True,
+        enable_same_dimensions_track_constraint: bool = False,
+        enable_save_button: bool = False,
         organization: Optional[str] = None,
     ) -> Dataset:
         """Add a dataset.
@@ -367,6 +374,8 @@ class SegmentsClient:
             enable_skip_reviewing: Enable the skip button in the reviewing workflow. Defaults to :obj:`False`.
             enable_ratings: Enable star-ratings for labeled images. Defaults to :obj:`False`.
             enable_interpolation: Enable label interpolation in sequence datasets. Ignored for non-sequence datasets. Defaults to :obj:`True`.
+            enable_same_dimensions_track_constraint: Enable constraint to keep same cuboid dimensions for the entire object track in point cloud cuboid datasets. Ignored for non-cuboid datasets. Defaults to :obj:`False`.
+            enable_save_button: Enable a save button in the labeling and reviewing workflow, to save unfinished work. Defaults to :obj:`False`.
             organization: The username of the organization for which this dataset should be created. None will create a dataset for the current user. Defaults to :obj:`None`.
         Raises:
             :exc:`~segments.exceptions.ValidationError`: If validation of the task attributes fails.
@@ -383,7 +392,7 @@ class SegmentsClient:
                 "categories": [{"id": 1, "name": "object"}],
             }
 
-        if type(task_attributes) is dict and not self.disable_typechecking:
+        if type(task_attributes) is dict:
             try:
                 TaskAttributes.parse_obj(task_attributes)
             except pydantic.ValidationError as e:
@@ -406,6 +415,8 @@ class SegmentsClient:
             "enable_skip_reviewing": enable_skip_reviewing,
             "enable_ratings": enable_ratings,
             "enable_interpolation": enable_interpolation,
+            "enable_same_dimensions_track_constraint": enable_same_dimensions_track_constraint,
+            "enable_save_button": enable_save_button,
             "data_type": "IMAGE",
         }
 
@@ -432,6 +443,8 @@ class SegmentsClient:
         enable_skip_reviewing: Optional[bool] = None,
         enable_ratings: Optional[bool] = None,
         enable_interpolation: Optional[bool] = None,
+        enable_same_dimensions_track_constraint: Optional[bool] = None,
+        enable_save_button: Optional[bool] = None,
     ) -> Dataset:
         """Update a dataset.
 
@@ -454,6 +467,8 @@ class SegmentsClient:
             enable_skip_reviewing: Enable the skip button in the reviewing workflow. Defaults to :obj:`None`.
             enable_ratings: Enable star-ratings for labeled images. Defaults to :obj:`None`.
             enable_interpolation: Enable label interpolation in sequence datasets. Ignored for non-sequence datasets. Defaults to :obj:`None`.
+            enable_same_dimensions_track_constraint: Enable constraint to keep same cuboid dimensions for the entire object track in point cloud cuboid datasets. Ignored for non-cuboid datasets. Defaults to :obj:`None`.
+            enable_save_button: Enable a save button in the labeling and reviewing workflow, to save unfinished work. Defaults to :obj:`False`.
         Raises:
             :exc:`~segments.exceptions.ValidationError`: If validation of the dataset fails.
             :exc:`~segments.exceptions.APILimitError`: If the API limit is exceeded.
@@ -498,6 +513,14 @@ class SegmentsClient:
         if enable_interpolation:
             payload["enable_interpolation"] = enable_interpolation
 
+        if enable_same_dimensions_track_constraint:
+            payload[
+                "enable_same_dimensions_track_constraint"
+            ] = enable_same_dimensions_track_constraint
+
+        if enable_save_button:
+            payload["enable_save_button"] = enable_save_button
+
         r = self._patch(f"/datasets/{dataset_identifier}/", data=payload, model=Dataset)
         # logger.info(f"Updated {dataset_identifier}")
 
@@ -770,11 +793,10 @@ class SegmentsClient:
         for result in results:
             result.pop("label", None)
 
-        if not self.disable_typechecking:
-            try:
-                results = parse_obj_as(List[Sample], results)
-            except pydantic.ValidationError as e:
-                raise ValidationError(message=str(e), cause=e)
+        try:
+            results = parse_obj_as(List[Sample], results)
+        except pydantic.ValidationError as e:
+            raise ValidationError(message=str(e), cause=e)
 
         return cast(List[Sample], results)
 
@@ -814,6 +836,8 @@ class SegmentsClient:
         attributes: Union[Dict[str, Any], SampleAttributes],
         metadata: Optional[Dict[str, Any]] = None,
         priority: float = 0,
+        assigned_labeler: Optional[str] = None,
+        assigned_reviewer: Optional[str] = None,
         embedding: Optional[Union[npt.NDArray[Any], List[float]]] = None,
     ) -> Sample:
         """Add a sample to a dataset.
@@ -848,6 +872,8 @@ class SegmentsClient:
             attributes: The sample attributes. Please refer to the `online documentation <https://docs.segments.ai/reference/sample-and-label-types/sample-types>`__.
             metadata: Any sample metadata. Example: ``{'weather': 'sunny', 'camera_id': 3}``.
             priority: Priority in the labeling queue. Samples with higher values will be labeled first. Defaults to ``0``.
+            assigned_labeler: The username of the user who should label this sample. Leave empty to not assign a specific labeler. Defaults to :obj:`None`.
+            assigned_reviewer: The username of the user who should review this sample. Leave empty to not assign a specific reviewer. Defaults to :obj:`None`.
             embedding: Embedding of this sample represented by an array of floats.
         Raises:
             :exc:`~segments.exceptions.ValidationError`: If validation of the sample attributes fails.
@@ -859,7 +885,7 @@ class SegmentsClient:
             :exc:`~segments.exceptions.TimeoutError`: If the request times out.
         """
 
-        if type(attributes) is dict and not self.disable_typechecking:
+        if type(attributes) is dict:
             try:
                 parse_obj_as(SampleAttributes, attributes)
             except pydantic.ValidationError as e:
@@ -867,7 +893,7 @@ class SegmentsClient:
                     "Did you use the right sample attributes? Please refer to the online documentation: https://docs.segments.ai/reference/sample-and-label-types/sample-types.",
                 )
                 raise ValidationError(message=str(e), cause=e)
-        elif type(attributes) is SampleAttributes:
+        elif type(attributes) in get_args(SampleAttributes):
             attributes = attributes.dict()
 
         payload: Dict[str, Any] = {
@@ -881,6 +907,12 @@ class SegmentsClient:
         if priority:
             payload["priority"] = priority
 
+        if assigned_labeler:
+            payload["assigned_labeler"] = assigned_labeler
+
+        if assigned_reviewer:
+            payload["assigned_reviewer"] = assigned_reviewer
+
         if embedding:
             payload["embedding"] = embedding
 
@@ -914,7 +946,7 @@ class SegmentsClient:
 
         # Check the input
         for sample in samples:
-            if type(sample) is dict and not self.disable_typechecking:
+            if type(sample) is dict:
                 if "name" not in sample or "attributes" not in sample:
                     raise KeyError(
                         f"Please add a name and attributes to your sample: {sample}"
@@ -946,7 +978,9 @@ class SegmentsClient:
         name: Optional[str] = None,
         attributes: Optional[Union[Dict[str, Any], SampleAttributes]] = None,
         metadata: Optional[Dict[str, Any]] = None,
-        priority: float = 0,
+        priority: Optional[float] = None,
+        assigned_labeler: Optional[str] = None,
+        assigned_reviewer: Optional[str] = None,
         embedding: Optional[Union[npt.NDArray[Any], List[float]]] = None,
     ) -> Sample:
         """Update a sample.
@@ -968,7 +1002,9 @@ class SegmentsClient:
             name: The name of the sample.
             attributes: The sample attributes. Please refer to the `online documentation <https://docs.segments.ai/reference/sample-and-label-types/sample-types>`__.
             metadata: Any sample metadata. Example: ``{'weather': 'sunny', 'camera_id': 3}``.
-            priority: Priority in the labeling queue. Samples with higher values will be labeled first. Default is ``0``.
+            priority: Priority in the labeling queue. Samples with higher values will be labeled first.
+            assigned_labeler: The username of the user who should label this sample. Leave empty to not assign a specific labeler.
+            assigned_reviewer: The username of the user who should review this sample. Leave empty to not assign a specific reviewer.
             embedding: Embedding of this sample represented by list of floats.
         Raises:
             :exc:`~segments.exceptions.APILimitError`: If the API limit is exceeded.
@@ -986,7 +1022,7 @@ class SegmentsClient:
         if attributes:
             payload["attributes"] = (
                 attributes.dict()
-                if type(attributes) is SampleAttributes
+                if type(attributes) in get_args(SampleAttributes)
                 else attributes
             )
 
@@ -996,6 +1032,12 @@ class SegmentsClient:
         if priority:
             payload["priority"] = priority
 
+        if assigned_labeler:
+            payload["assigned_labeler"] = assigned_labeler
+
+        if assigned_reviewer:
+            payload["assigned_reviewer"] = assigned_reviewer
+
         if embedding:
             payload["embedding"] = embedding
 
@@ -1099,7 +1141,7 @@ class SegmentsClient:
             :exc:`~segments.exceptions.TimeoutError`: If the request times out.
         """
 
-        if type(attributes) is dict and not self.disable_typechecking:
+        if type(attributes) is dict:
             try:
                 parse_obj_as(LabelAttributes, attributes)
             except pydantic.ValidationError as e:
@@ -1107,7 +1149,7 @@ class SegmentsClient:
                     "Did you use the right label attributes? Please refer to the online documentation: https://docs.segments.ai/reference/sample-and-label-types/label-types.",
                 )
                 raise ValidationError(message=str(e), cause=e)
-        elif type(attributes) is LabelAttributes:
+        elif type(attributes) in get_args(LabelAttributes):
             attributes = attributes.dict()
 
         payload: Dict[str, Any] = {
@@ -1169,7 +1211,9 @@ class SegmentsClient:
 
         if attributes:
             payload["attributes"] = (
-                attributes.dict() if type(attributes) is LabelAttributes else attributes
+                attributes.dict()
+                if type(attributes) in get_args(LabelAttributes)
+                else attributes
             )
 
         if label_status:
@@ -1566,7 +1610,7 @@ class SegmentsClient:
     ####################
     # Helper functions #
     ####################
-    @exception_handler
+    @handle_exceptions
     def _get(
         self,
         endpoint: str,
@@ -1594,7 +1638,7 @@ class SegmentsClient:
 
         return r
 
-    @exception_handler
+    @handle_exceptions
     def _post(
         self,
         endpoint: str,
@@ -1625,7 +1669,7 @@ class SegmentsClient:
 
         return r
 
-    @exception_handler
+    @handle_exceptions
     def _put(
         self,
         endpoint: str,
@@ -1656,7 +1700,7 @@ class SegmentsClient:
 
         return r
 
-    @exception_handler
+    @handle_exceptions
     def _patch(
         self,
         endpoint: str,
@@ -1687,7 +1731,7 @@ class SegmentsClient:
 
         return r
 
-    @exception_handler
+    @handle_exceptions
     def _delete(
         self,
         endpoint: str,
@@ -1722,7 +1766,7 @@ class SegmentsClient:
         """Get the authorization header with the API key."""
         return {"Authorization": f"APIKey {self.api_key}"} if self.api_key else None
 
-    @exception_handler
+    @handle_exceptions
     def _upload_to_aws(
         self, file: Union[TextIO, BinaryIO], url: str, aws_fields: AWSFields
     ) -> requests.Response:
