@@ -1,15 +1,21 @@
+from __future__ import annotations
+
 from typing import Any, Dict, List, Optional, Tuple, Union
 
 from pydantic import BaseModel as PydanticBaseModel
-from pydantic import Extra, validator
+from pydantic import validator
 from segments.exceptions import ValidationError
-from typing_extensions import Literal, TypedDict
+from typing_extensions import Literal, TypedDict, get_args
 
 
 class BaseModel(PydanticBaseModel):
     class Config:
-        # https://pydantic-docs.helpmanual.io/usage/model_config/#smart-union
+        # Smart union checks all types before deciding which one to pick (otherwise, first type that fits). https://pydantic-docs.helpmanual.io/usage/model_config/#smart-union
         smart_union = True
+        # What happens with extra fields in dictionaries. Use ignore in production and allow in debug mode. https://pydantic-docs.helpmanual.io/usage/model_config/#change-behaviour-globally
+        extra = "ignore"
+        # What happens with wrong field types. Use false in production and true in debug mode. https://pydantic-docs.helpmanual.io/usage/types/#arbitrary-types-allowed
+        arbitrary_types_allowed = False
 
 
 #######################################
@@ -35,19 +41,15 @@ TaskType = Literal[
     "pointcloud-cuboid-sequence",
     "pointcloud-segmentation",
     "pointcloud-segmentation-sequence",
+    "pointcloud-vector",
+    "pointcloud-vector-sequence",
+    "text-named-entities",
     "text-named-entities",
     "text-span-categorization",
     "",
 ]
-DataType = Literal["IMAGE"]
 Role = Literal["labeler", "reviewer", "admin"]
-Status = Literal["PENDING", "SUCCEEDED", "FAILED"]
 IssueStatus = Literal["OPEN", "CLOSED"]
-ReleaseType = Literal["JSON"]
-ImageVectorAnnotationType = Literal["bbox", "polygon", "polyline", "point"]
-PointcloudAnnotationType = Literal["cuboid"]
-PCDType = Literal["pcd", "kitti", "nuscenes"]
-InputType = Literal["select", "text", "number", "checkbox"]
 Category = Literal[
     "street_scenery",
     "garden",
@@ -58,17 +60,6 @@ Category = Literal[
     "fruit",
     "other",
 ]
-_Category = [
-    "street_scenery",
-    "garden",
-    "agriculture",
-    "satellite",
-    "people",
-    "medical",
-    "fruit",
-    "other",
-]
-DistortionModel = Literal["plumb_bob"]
 RGB = Tuple[int, int, int]
 RGBA = Tuple[int, int, int, int]
 FormatVersion = Union[float, str]
@@ -86,16 +77,16 @@ class AuthHeader(TypedDict):
 class URL(BaseModel):
     url: Optional[
         str
-    ] = None  # TODO Remove optional (e.g., the backend does not return an URL when adding a release).
+    ]  # TODO Remove optional (e.g., the backend does not return an URL when adding a release).
 
 
 class Release(BaseModel):
     uuid: str
     name: str
     description: str
-    release_type: ReleaseType
+    release_type: Literal["JSON"]
     attributes: URL
-    status: Status
+    status: Literal["PENDING", "SUCCEEDED", "FAILED"]
     # status_info: str
     created_at: str
     samples_count: int
@@ -161,15 +152,15 @@ class File(BaseModel):
 class Annotation(BaseModel):
     id: int
     category_id: int
-    attributes: Optional[ObjectAttributes] = None
+    attributes: Optional[ObjectAttributes]
 
 
 # Image segmenation
 class ImageSegmentationLabelAttributes(BaseModel):
     annotations: List[Annotation]
     segmentation_bitmap: URL
-    image_attributes: Optional[ImageAttributes] = None
-    format_version: Optional[FormatVersion] = None
+    image_attributes: Optional[ImageAttributes]
+    format_version: Optional[FormatVersion]
 
 
 # Image vector
@@ -178,14 +169,14 @@ class ImageVectorAnnotation(BaseModel):
     id: int
     category_id: int
     points: List[List[float]]
-    type: ImageVectorAnnotationType
-    attributes: Optional[ObjectAttributes] = None
+    type: Literal["bbox", "polygon", "polyline", "point"]
+    attributes: Optional[ObjectAttributes]
 
 
 class ImageVectorLabelAttributes(BaseModel):
     annotations: List[ImageVectorAnnotation]
-    format_version: Optional[FormatVersion] = None
-    image_attributes: Optional[ImageAttributes] = None
+    format_version: Optional[FormatVersion]
+    image_attributes: Optional[ImageAttributes]
 
 
 # Image sequence vector
@@ -196,21 +187,21 @@ class ImageSequenceVectorAnnotation(ImageVectorAnnotation):
 
 class ImageVectorFrame(BaseModel):
     annotations: List[ImageSequenceVectorAnnotation]
-    timestamp: Optional[int] = None
-    format_version: Optional[FormatVersion] = None
-    image_attributes: Optional[ImageAttributes] = None
+    timestamp: Optional[str]
+    format_version: Optional[FormatVersion]
+    image_attributes: Optional[ImageAttributes]
 
 
 class ImageSequenceVectorLabelAttributes(BaseModel):
     frames: List[ImageVectorFrame]
-    format_version: Optional[FormatVersion] = None
+    format_version: Optional[FormatVersion]
 
 
 # Point cloud segmentation
 class PointcloudSegmentationLabelAttributes(BaseModel):
     annotations: List[Annotation]
     point_annotations: List[int]
-    format_version: Optional[FormatVersion] = None
+    format_version: Optional[FormatVersion]
 
 
 class XYZ(BaseModel):
@@ -227,13 +218,27 @@ class PointcloudCuboidAnnotation(BaseModel):
     position: XYZ
     dimensions: XYZ
     yaw: float
-    type: PointcloudAnnotationType
-    attributes: Optional[ObjectAttributes] = None
+    type: Literal["cuboid", "cuboid-sync"]
+    attributes: Optional[ObjectAttributes]
 
 
 class PointcloudCuboidLabelAttributes(BaseModel):
     annotations: List[PointcloudCuboidAnnotation]
-    format_version: Optional[FormatVersion] = None
+    format_version: Optional[FormatVersion]
+
+
+# Point cloud vector
+class PointcloudVectorAnnotation(BaseModel):
+    id: int
+    category_id: int
+    points: List[List[float]]
+    type: Literal["polygon", "polyline", "point"]
+    attributes: Optional[ObjectAttributes]
+
+
+class PointcloudVectorLabelAttributes(BaseModel):
+    annotations: List[PointcloudVectorAnnotation]
+    format_version: Optional[FormatVersion]
 
 
 # Point cloud sequence segmentation
@@ -242,18 +247,19 @@ class PointcloudSequenceSegmentationAnnotation(BaseModel):
     category_id: int
     track_id: int
     is_keyframe: bool = False
-    attributes: Optional[ObjectAttributes] = None
+    attributes: Optional[ObjectAttributes]
 
 
 class PointcloudSegmentationFrame(BaseModel):
     annotations: List[PointcloudSequenceSegmentationAnnotation]
-    point_annotations: Optional[List[int]] = None
-    format_version: Optional[FormatVersion] = None
+    point_annotations: Optional[List[int]]
+    timestamp: Optional[str]
+    format_version: Optional[FormatVersion]
 
 
 class PointcloudSequenceSegmentationLabelAttributes(BaseModel):
     frames: List[PointcloudSegmentationFrame]
-    format_version: Optional[FormatVersion] = None
+    format_version: Optional[FormatVersion]
 
 
 # Point cloud sequence cuboid
@@ -263,14 +269,31 @@ class PointcloudSequenceCuboidAnnotation(PointcloudCuboidAnnotation):
 
 
 class PointcloudSequenceCuboidFrame(BaseModel):
-    timestamp: int
     annotations: List[PointcloudSequenceCuboidAnnotation]
-    format_version: Optional[FormatVersion] = None
+    timestamp: Optional[str]
+    format_version: Optional[FormatVersion]
 
 
 class PointcloudSequenceCuboidLabelAttributes(BaseModel):
     frames: List[PointcloudSequenceCuboidFrame]
-    format_version: Optional[FormatVersion] = None
+    format_version: Optional[FormatVersion]
+
+
+# Point cloud sequence vector
+class PointcloudSequenceVectorAnnotation(PointcloudVectorAnnotation):
+    track_id: int
+    is_keyframe: bool = False
+
+
+class PointcloudSequenceVectorFrame(BaseModel):
+    annotations: List[PointcloudSequenceVectorAnnotation]
+    format_version: Optional[FormatVersion]
+    timestamp: Optional[str]
+
+
+class PointcloudSequenceVectorLabelAttributes(BaseModel):
+    frames: List[PointcloudSequenceVectorFrame]
+    format_version: Optional[FormatVersion]
 
 
 # Text
@@ -282,17 +305,21 @@ class TextAnnotation(BaseModel):
 
 class TextLabelAttributes(BaseModel):
     annotations: List[TextAnnotation]
-    format_version: Optional[FormatVersion] = None
+    format_version: Optional[FormatVersion]
 
 
+# Most specific type first
+# https://pydantic-docs.helpmanual.io/usage/types/#unions
 LabelAttributes = Union[
-    ImageSegmentationLabelAttributes,
     ImageVectorLabelAttributes,
+    ImageSegmentationLabelAttributes,
     ImageSequenceVectorLabelAttributes,
-    PointcloudSegmentationLabelAttributes,
     PointcloudCuboidLabelAttributes,
-    PointcloudSequenceSegmentationLabelAttributes,
+    PointcloudVectorLabelAttributes,
+    PointcloudSegmentationLabelAttributes,
     PointcloudSequenceCuboidLabelAttributes,
+    PointcloudSequenceVectorLabelAttributes,
+    PointcloudSequenceSegmentationLabelAttributes,
     TextLabelAttributes,
 ]
 
@@ -306,10 +333,10 @@ class Label(BaseModel):
     created_at: str
     created_by: str
     updated_at: str
-    score: Optional[float] = None
-    rating: Optional[float] = None
-    reviewed_at: Optional[str] = None
-    reviewed_by: Optional[str] = None
+    score: Optional[float]
+    rating: Optional[float]
+    reviewed_at: Optional[str]
+    reviewed_by: Optional[str]
 
 
 ##########
@@ -322,7 +349,7 @@ class ImageSampleAttributes(BaseModel):
 
 # Image sequence
 class ImageFrame(ImageSampleAttributes):
-    name: Optional[str] = None
+    name: Optional[str]
 
 
 class ImageSequenceSampleAttributes(BaseModel):
@@ -332,7 +359,7 @@ class ImageSequenceSampleAttributes(BaseModel):
 # Point cloud
 class PCD(BaseModel):
     url: str
-    type: PCDType = "pcd"
+    type: Literal["pcd", "kitti", "nuscenes"] = "pcd"
 
 
 class XYZW(BaseModel):
@@ -359,17 +386,17 @@ class CameraExtrinsics(BaseModel):
 class CalibratedImage(URL):
     row: int
     col: int
-    intrinsics: Optional[CameraIntrinsics] = None
-    extrinsics: Optional[CameraExtrinsics] = None
+    intrinsics: Optional[CameraIntrinsics]
+    extrinsics: Optional[CameraExtrinsics]
 
 
 class PointcloudSampleAttributes(BaseModel):
     pcd: PCD
-    images: Optional[List[CalibratedImage]] = None
-    ego_pose: Optional[EgoPose] = None
-    default_z: Optional[float] = None
-    name: Optional[str] = None
-    timestamp: Optional[int] = None
+    images: Optional[List[CalibratedImage]]
+    ego_pose: Optional[EgoPose]
+    default_z: Optional[float]
+    name: Optional[str]
+    timestamp: Optional[str]
 
 
 # Point cloud sequence
@@ -398,6 +425,8 @@ class Sample(BaseModel):
     metadata: Dict[str, Any]
     created_at: str
     created_by: str
+    assigned_labeler: Optional[str]
+    assigned_reviewer: Optional[str]
     comments: Optional[List[str]]
     priority: float
     has_embedding: Optional[bool]
@@ -423,28 +452,28 @@ class SelectTaskAttribute(BaseModel):
     name: str
     input_type: Literal["select"]
     values: List[str]
-    default_value: Optional[str] = None
+    default_value: Optional[str]
 
 
 class TextTaskAttribute(BaseModel):
     name: str
     input_type: Literal["text"]
-    default_value: Optional[str] = None
+    default_value: Optional[str]
 
 
 class NumberTaskAttribute(BaseModel):
     name: str
     input_type: Literal["number"]
-    default_value: Optional[float] = None
-    min: Optional[float] = None
-    max: Optional[float] = None
-    step: Optional[float] = None
+    default_value: Optional[float]
+    min: Optional[float]
+    max: Optional[float]
+    step: Optional[float]
 
 
 class CheckboxTaskAttribute(BaseModel):
     name: str
     input_type: Literal["checkbox"]
-    default_value: Optional[bool] = None
+    default_value: Optional[bool]
 
 
 TaskAttribute = Union[
@@ -458,23 +487,28 @@ TaskAttribute = Union[
 class TaskAttributeCategory(BaseModel):
     name: str
     id: int
-    color: Optional[Union[RGB, RGBA]] = None
-    attributes: Optional[List[TaskAttribute]] = None
-    dimensions: Optional[XYZ] = None
+    color: Optional[Union[RGB, RGBA]]
+    has_instances: Optional[bool]
+    attributes: Optional[List[TaskAttribute]]
+    dimensions: Optional[XYZ]
+
+    class Config:
+        extra = "allow"
 
 
 class TaskAttributes(BaseModel):
-    format_version: Optional[FormatVersion] = None
-    categories: Optional[List[TaskAttributeCategory]] = None
+    format_version: Optional[FormatVersion]
+    categories: Optional[List[TaskAttributeCategory]]
+    image_attributes: Optional[List[TaskAttribute]]
 
     class Config:
-        extra = Extra.allow
+        extra = "allow"
 
 
 class Owner(BaseModel):
     username: str
     created_at: str
-    email: Optional[str] = None
+    email: Optional[str]
 
 
 class Statistics(BaseModel):
@@ -489,29 +523,29 @@ class Statistics(BaseModel):
 class Labelset(BaseModel):
     name: str
     description: str
-    # uuid: Optional[str] = None
-    # readme: Optional[str] = None
-    # task_type: Optional[TaskType] = None
-    # attributes: Optional[Union[str, TaskAttributes]] = None
-    is_groundtruth: Optional[bool] = None
-    # statistics: Optional[Statistics] = None
-    created_at: Optional[str] = None
-    # stats: Optional[Dict[str, Any]] = None
+    # uuid: Optional[str]
+    # readme: Optional[str]
+    # task_type: Optional[TaskType]
+    # attributes: Optional[Union[str, TaskAttributes]]
+    is_groundtruth: Optional[bool]
+    # statistics: Optional[Statistics]
+    created_at: Optional[str]
+    # stats: Optional[Dict[str, Any]]
 
 
 class LabelStats(BaseModel):
-    TOTAL: Optional[int] = None
-    LABELED: Optional[int] = None
-    UNLABELED: Optional[int] = None
-    PRELABELED: Optional[int] = None
+    TOTAL: Optional[int]
+    LABELED: Optional[int]
+    UNLABELED: Optional[int]
+    PRELABELED: Optional[int]
 
 
 class Dataset(BaseModel):
     name: str
     full_name: str
-    cloned_from: Optional[str] = None
+    cloned_from: Optional[str]
     description: str
-    # data_type: DataType
+    # data_type: Literal["IMAGE"]
     category: str  # Category
     public: bool
     owner: Owner
@@ -525,22 +559,23 @@ class Dataset(BaseModel):
     task_type: TaskType
     # task_readme: str
     label_stats: LabelStats
-    samples_count: Optional[Union[str, int]] = None
-    collaborators_count: Optional[int] = None
-    task_attributes: Optional[TaskAttributes] = None
-    labelsets: Optional[List[Labelset]] = None
-    role: Optional[str] = None
-    readme: Optional[str] = None
-    noncollaborator_can_label: Optional[bool] = None
-    noncollaborator_can_review: Optional[bool] = None
-    # tasks: Optional[List[Dict[str, Any]]] = None
-    embeddings_enabled: Optional[bool] = None
+    samples_count: Optional[Union[str, int]]
+    collaborators_count: Optional[int]
+    task_attributes: Optional[TaskAttributes]
+    labelsets: Optional[List[Labelset]]
+    role: Optional[str]
+    readme: Optional[str]
+    noncollaborator_can_label: Optional[bool]
+    noncollaborator_can_review: Optional[bool]
+    # tasks: Optional[List[Dict[str, Any]]]
+    embeddings_enabled: Optional[bool]
 
     @validator("category")
     def check_category(cls, category: str) -> str:
-        if category not in _Category and "custom-" not in category:
+        category_list = get_args(Category)
+        if category not in category_list and "custom-" not in category:
             raise ValidationError(
-                f"The category should be one of {_Category}, but is {category}."
+                f"The category should be one of {category_list}, but is {category}."
             )
         return category
 
@@ -551,9 +586,9 @@ class Dataset(BaseModel):
 class SegmentsDatasetCategory(BaseModel):
     id: int
     name: str
-    color: Optional[Union[RGB, RGBA]] = None
-    attributes: Optional[List[Any]] = None
+    color: Optional[Union[RGB, RGBA]]
+    attributes: Optional[List[Any]]
 
     class Config:
         allow_population_by_field_name = True
-        extra = Extra.allow
+        extra = "allow"
